;
; DESCRIPT.ASM
;
; This file contains a set of USB descriptors for the device.  
;
; The device has four buttons and two LEDs, but that doesn't really correspond to any
; real-world device.  To avoid problems with things like Windows, we therefore pretend
; to be a Joystick, whose minimal definition is two analogue axes and two buttons.  In our
; case we fake up two basic axes that never change, and increase the number of
; buttons.  This is the approach taken by the Saitek P8000, for example.
;

#include <p16C745.inc>
#include "usb_defs.inc"

lohi	macro	arg
	retlw	low arg
	retlw	high arg
	endm

;
; Some useful macros for describing HID report items.
;
hid_item	macro	tag, arg
	retlw	tag
	retlw	arg
	endm

hid_usage_page	macro	arg
	hid_item	0x05, arg
	endm

hid_usage	macro	arg
	hid_item	0x09, arg
	endm

hid_collection	macro	arg
	hid_item	0xa1, arg
	endm

hid_collection_end	macro
	retlw	0xc0
	endm

hid_usage_min	macro	arg
	hid_item	0x19, arg
	endm

hid_usage_max	macro	arg
	hid_item	0x29, arg
	endm

hid_rep_count	macro	arg
	hid_item	0x95, arg
	endm

hid_rep_size	macro	arg
	hid_item	0x75, arg
	endm

hid_log_min	macro	arg
	hid_item	0x15, arg
	endm

hid_log_max	macro	arg
	hid_item	0x25, arg
	endm

hid_input	macro	arg
	hid_item	0x81, arg
	endm

hid_output	macro	arg
	hid_item	0x91, arg
	endm

USBBANK code
	global  Config_desc_index
	global  Report_desc_index
	global  Descriptions
	global  string_index
	global  DeviceDescriptor
	global  ReportDescriptor
	global  ReportDescriptorLen
	global  String0
	global  String0_end
	global  StringDescriptions
	global  HID_Descriptor

	extern  EP0_start
	extern  temp    ; temp var used in get config index
	extern  temp2   ; another temp, in bank2 

; ******************************************************************
; Given a configuration descriptor index, returns the beginning address
; of the descriptor within the descriptions table
; ******************************************************************
Config_desc_index
	movwf   temp
	movlw   HIGH CDI_start
	movwf   PCLATH
	movlw   low  CDI_start
	addwf   temp,w
	btfsc   STATUS,C
	incf    PCLATH,f
	movwf   PCL
CDI_start           ; this table calculates the offsets for each configuration
	retlw   low  Config1       ; descriptor from the beginning
	retlw   high Config1       ; of the table, effectively
	; more configurations can be added here
	; retlw   low Config2
	; retlw   high Config2
	; etc....

; ******************************************************************
; Given a report descriptor index, returns the beginning address
; of the descriptor within the descriptions table
; ******************************************************************
Report_desc_index
	movwf   temp
	movlw   HIGH RDI_start
	movwf   PCLATH
	movlw   low  RDI_start
	addwf   temp,w
	btfsc   STATUS,C
	incf    PCLATH,f
	movwf   PCL
; this table calculates the offsets for each report
; descriptor from the beginning
; of the table, effectively
RDI_start
	lohi	ReportDescriptorLen
	; more reports can be added here
	; lohi  ReportDescriptorLen2
	; etc....

; ******************************************************************
; This table is polled by the host immediately after USB Reset has been released.
; This table defines the maximum packet size EP0 can take.
; See section 9.6.1 of the Rev 1.0 USB specification.
; These fields are application DEPENDENT. Modify these to meet
; your specifications.
; the offset is passed in P0 and P1 (P0 is low order byte).
; ******************************************************************
Descriptions
	banksel EP0_start
	movf    EP0_start+1,w
	movwf   PCLATH
	movf    EP0_start,w
	movwf   PCL

DeviceDescriptor
StartDevDescr
	retlw	0x12    ; bLength	Length of this descriptor
	retlw	DEVICE	; bDescType
	lohi	0x0110	; bcdUSB	USB revision 1.10
	retlw	0x00    ; bDeviceClass	zero means each interface operates independently
	retlw	0x00    ; bDeviceSubClass
	retlw	0x00    ; bDeviceProtocol
	retlw	0x08	; bMaxPacketSize0 - inited in UsbInit()
	lohi	0x04D8	; idVendor	Microchip
	lohi	0x0003	; ifProduct	TB054 example
	lohi	0x0007	; bcdDevice	device release number (BCD)
	retlw	0x01	; iManufacturer
	retlw	0x02	; iProduct
	; somehow, changing iSerialNumber will cause windows to display
	; "New Hardware found" dialog. The device will function normally,
	; however, USBCheck will NOT start. From CATC we can see that
	; the string is correctly transmitted.
	retlw	0;0x0	; iSerialNumber - 3
	retlw	NUM_CONFIGURATIONS	; bNumConfigurations

; ******************************************************************
; This table is retrieved by the host after the address has been set.
; This table defines the configurations available for the device.
; See section 9.6.2 of the Rev 1.0 USB specification (page 184).
; These fields are application DEPENDENT. 
; Modify these to meet your specifications.
; ******************************************************************
Config1
	retlw	0x09	; bLength	Length of this descriptor
	retlw	CONFIGURATION	; bDescType

	retlw	EndConfig1 - Config1
	retlw	0x00   
	retlw	0x01   	; bNumInterfaces	Number of interfaces

	retlw	0x01   	; bConfigValue	Configuration Value
	retlw	0x04   	; iConfig	String Index for this config = #01
	retlw	0x80	; bmAttributes	attributes - bus powered
	retlw	D'25'	; MaxPower	draws 50mA from the bus.

Interface1
	retlw	0x09	; length of descriptor
	retlw	INTERFACE
	retlw	0x00	; number of interface, 0 based array
	retlw	0x00	; alternate setting
	retlw	0x01	; number of endpoints used in this interface
	retlw	0x03	; interface class - assigned by the USB
	retlw	0x00	; boot device
	retlw	0x00	; interface protocol
	retlw 	0x05    ; index to string descriptor that describes this interface

HID_Descriptor
	retlw	0x09    ; descriptor size (9 bytes)
	retlw	0x21    ; descriptor type (HID)
	lohi	0x0100	; HID class specification release number (BCD)
	retlw	0x00    ; Localized country code (none)
	retlw	0x01    ; # of HID class descriptor to follow (1)
	retlw	0x22    ; Report descriptor type (HID)
	retlw	(end_ReportDescriptor - ReportDescriptor)
	retlw	0x00

	;
	; Descriptor for endpoint 1 IN
	;
	retlw	0x07    ; length of descriptor
	retlw	ENDPOINT
	retlw	0x81	; EP1, In
	retlw	0x03   	; Interrupt
	lohi	8	; max packet size (bytes)
	retlw	D'10'	; polling interval (ms)

	;
	; Descriptor for endpoint 1 OUT
	;
;	retlw	0x07	; length of descriptor
;	retlw	ENDPOINT
;	retlw	0x01	; EP1, Out
;	retlw	0x03	; Interrupt
;	lohi	8	; max packet size (bytes)
;	retlw	D'10'	; polling interval (ms)

EndConfig1

ReportDescriptorLen
	retlw	low  (end_ReportDescriptor-ReportDescriptor)

ReportDescriptor
	hid_usage_page	0x01	; USAGE_PAGE (Generic Desktop)
	hid_usage	0x04	; USAGE (Joystick)
	hid_collection	0x01	; COLLECTION (Application)

	;
	; Physical collection for the two axes.
	;
	hid_usage	0x01	;   USAGE(Pointer)
	hid_collection	0x00	;   COLLECTION(Physical)
	hid_usage	0x30	;      USAGE(X)
	hid_usage	0x31	;      USAGE(Y)
	hid_log_min	0	;      LOGICAL_MINIMUM (0)
	hid_log_max	0x7F	;      LOGICAL_MAXIMUM (127)
	hid_rep_size	8	;      REPORT_SIZE (8)
	hid_rep_count	2	;      REPORT_COUNT (2)
	hid_input	0x02	;      INPUT (Data,Var,Abs)
	hid_collection_end	;   END_COLLECTION !physical
	
	;
	; Describe the buttons.
	;
	hid_usage_page	0x09	;   USAGE_PAGE (Button)
	hid_usage_min	1	;   USAGE_MINIMUM (Button 1)
	hid_usage_max	4	;   USAGE_MAXIMUM (Button 4)
	hid_log_min	0	;   LOGICAL_MINIMUM (0)
	hid_log_max	1	;   LOGICAL_MAXIMUM (1)
	hid_rep_size	1	;   REPORT_SIZE (1)
	hid_rep_count	4	;   REPORT_COUNT (4)
	hid_input	0x02	;   INPUT (Data,Var,Abs)

	;
	; Describe enough padding bits to bring us up to the next byte boundary.
	;
	hid_rep_count	4	;   REPORT_COUNT (4)
	hid_input	0x01	;   INPUT (Constant)

	;
	; Describe the LEDs
	;
	hid_usage_page	0x08	;   USAGE_PAGE (LEDs)
	hid_usage_min	0x2C	;   USAGE_MINIMUM (busy)
	hid_usage_max	0x2D	;   USAGE_MAXIMUM (ready)
	hid_log_min	0	;   LOGICAL_MINIMUM (0)
	hid_log_max	1	;   LOGICAL_MAXIMUM (1)
	hid_rep_count	2	;   REPORT_COUNT (2)
	hid_rep_size	1	;   REPORT_SIZE (1)
	hid_output	0x02	;   OUTPUT (Data,Var,Abs)
	
	;
	; Describe enough padding bits to bring us up to the next byte boundary.
	;
	hid_rep_count	6	;   REPORT_COUNT (6)
	hid_output	0x03	;   OUTPUT (Constant,Var,Abs)

	hid_collection_end	; END_COLLECTION !application
end_ReportDescriptor

StringDescriptions
	banksel	EP0_start
	movf	EP0_start+1,w
	movwf	PCLATH
	movf	EP0_start,w
	movwf	PCL

; ******************************************************************
; Given a configuration descriptor index, returns the beginning address
; of the descriptor within the descriptions table
; ******************************************************************
string_index		; langid in W reg, string offset in EP0_start
	movwf	temp
	bcf	STATUS,C
	rlf	temp, f
	pagesel	langid_index
	call	langid_index
	movwf	temp2
	incf	temp, f
	pagesel	langid_index
	call	langid_index
	movwf	temp
	
	movf	temp, w
	movwf	PCLATH
	movf	temp2,w
	addwf	EP0_start+1,w
	btfsc	STATUS,C
	incf	PCLATH, f
	movwf	PCL

langid_index
	movlw	high langids
	movwf	PCLATH
	movlw	low langids
	addwf	temp, w
	btfsc	STATUS,C
	incf	PCLATH,f
	movwf	PCL

langids	; string indexes of different languages
	lohi	lang_1
	lohi	lang_2

lang_1				; English
	lohi	String0		; LangIDs
	lohi	String1_l1
	lohi	String2_l1
	lohi	String3_l1
	lohi	String4_l1
	lohi	String5_l1

lang_2				; Chinese...
	lohi	String0		; LangIDs
	lohi	String1_l2
	lohi	String2_l2
	lohi	String3_l2
	lohi	String4_l2
	lohi	String5_l2

String0
	retlw	low (String0_end - String0)	; length of string 
	retlw	STRING
	lohi	0x0409	; language ID (as defined by MS 0x0409)
	lohi	0x0804	; some other language ID for testing
String0_end

String1_l1
	retlw	String1_l1_end-String1_l1	; length of string
	retlw	STRING
	lohi	'I'
	lohi	'a'
	lohi	'n'
	lohi	' '
	lohi	'A'
	lohi	'.'
	lohi	' '
	lohi	'Y'
	lohi	'o'
	lohi	'u'
	lohi	'n'
	lohi	'g'
String1_l1_end

String2_l1
	retlw	String2_l1_end-String2_l1
	retlw	STRING
	lohi	'V'
	lohi	'o'
	lohi	't'
	lohi	'e'
	lohi	'r'
String2_l1_end

String3_l1
	retlw	String4_l1-String3_l1
	retlw	STRING
	lohi	'V'
	lohi	'1'
	lohi	'.'
	lohi	'1'
	lohi	'4'

String4_l1
	retlw	String5_l1-String4_l1
	retlw	STRING
	retlw	'C'
	retlw	0x00
	retlw	'f'
	retlw	0x00
	retlw 	'g'
	retlw	0x00
	retlw 	'1'
	retlw	0x00

String5_l1
	retlw	String6_l1-String5_l1
	retlw	STRING
	retlw	'E'
	retlw	0x00
	retlw	'P'
	retlw	0x00
	retlw	'1'
	retlw	0x00
	retlw	'0'
	retlw	0x00
	retlw	'I'
	retlw	0x00
	retlw	'n'
	retlw	0x00
String6_l1

String1_l2		; lang 2, chinese. String can be totally different than english
	retlw	String1_l2_end-String1_l2	; length of string
	retlw	STRING
	lohi	'I'
	lohi	'a'
	lohi	'n'
	lohi	' '
	lohi	'A'
	lohi	'.'
	lohi	' '
	lohi	'Y'
	lohi	'o'
	lohi	'u'
	lohi	'n'
	lohi	'g'
String1_l2_end

String2_l2
	retlw	String2_l2_end-String2_l2
	retlw	STRING
	lohi	'V'
	lohi	'o'
	lohi	't'
	lohi	'e'
	lohi	'r'
String2_l2_end

String3_l2
	retlw	String4_l2-String3_l2
	retlw	STRING
	retlw	'B'
	retlw	0x00
	retlw	'e'
	retlw	0x00
	retlw	't'
	retlw	0x00
	retlw	'a'
	retlw	0x00

String4_l2
	retlw	String5_l2-String4_l2
	retlw	STRING
	retlw	'C'
	retlw	0x00
	retlw	'f'
	retlw	0x00
	retlw 	'g'
	retlw	0x00
	retlw 	'1'
	retlw	0x00

String5_l2
	retlw	String6_l2-String5_l2
	retlw	STRING
	retlw	'E'
	retlw	0x00
	retlw	'P'
	retlw	0x00
	retlw	'1'
	retlw	0x00
	retlw	'0'
	retlw	0x00
	retlw	'I'
	retlw	0x00
	retlw	'n'
	retlw	0x00
String6_l2

    end
