;
; GAMEPAD.ASM
;
; Application code.

#include <p16c745.inc>
#include "usb_defs.inc"

	__CONFIG	_H4_OSC & _WDT_OFF & _PWRTE_OFF & _CP_OFF

unbanked	udata_shr
W_save		res	1

bank0		udata
Status_save	res	1	; registers for saving context
PCLATH_save	res	1
FSR_save	res	1
TEMP            res	1	; temporary registor
BUTTONS         res	1	; button states
LEDS		res	1	; lamp states
BUFFER		res	8	; source/destination buffer
FLAG            res	1	; contains flag bits

; FLAG bits
#define CHANGED FLAG,0  ; set if gamepad buttons have changed

	extern	InitUSB
	extern	PutEP1
	extern	GetEP1
	extern	ServiceUSBInt

STARTUP	code
	pagesel Main
	goto	Main
	nop

InterruptServiceVector
	movwf	W_save		; save W
	movf	STATUS,W
	clrf	STATUS		; force to page 0
	movwf	Status_save	; save STATUS
	movf	PCLATH,w
	movwf	PCLATH_save	; save PCLATH
	movf	FSR,w
	movwf	FSR_save	; save FSR

;*******************************************************************
; Interrupt Service Routine
; First we step through several stages, attempting to identify the source 
; of the interrupt.
;*******************************************************************

Process_ISR
; Step 1, what triggered the interrupt?
	banksel PIR1
	pagesel	ServiceUSBInt
	btfsc	PIR1,USBIF      ; If it was the USB interrupt flag then
	call	ServiceUSBInt   ;  goto its service routine.

;*******************************************************************
; End ISR, restore context and return to the main program
;*******************************************************************
EndISR
	clrf	STATUS		    ; select bank 0
	movf	FSR_save,w	    ; restore the FSR
	movwf	FSR
	movf	PCLATH_save,w	; restore PCLATH
	movwf	PCLATH
	movf	Status_save,w	; restore Status
	movwf	STATUS
	swapf	W_save,f    	; restore W without corrupting STATUS
	swapf	W_save,w
	retfie

	code

;*******************************************************************
;Main
; Initializes appropiate registers and calls the enumeration process
; for the USB.    
;*******************************************************************
Main
	movlw	.30	    	; delay 16 uS to wait for USB to reset
	movwf	W_save		; SIE before initializing registers
	decfsz	W_save,f	; W_save is merely a convienient register
	goto	$-1		; to use for the delay counter.
    
	call    InitGameport	; Initialize the registers for the mouse

	banksel	PORTA		; light first light after restart
	movlw	0x01
	movwf	PORTA

	pagesel	InitUSB
	call	InitUSB		; Initialize enumeration
	pagesel	Main
	ConfiguredUSB		; wait here until the enumeration
				;   process is complete

	banksel	PORTA		; light second light after config
	movlw	0x02
	movwf	PORTA

	banksel BUFFER

;
; Loop
;
; Main application loop:
;
;    * polls the buttons and sends their values to the host
;
;    * data arriving from the host goes to the LEDs
;
Loop
	;
	; Send button states to the host, if they have changed.
	;
	call    ReadButtons ; Read button states to w

	;
	; Set CHANGED if new value is different
	;
	xorwf   BUFFER+2,f
	btfss   STATUS,Z
	bsf     CHANGED

	;
	; Store new buttons value
	;
	movwf   BUFFER+2

	;
	; Fill two axis slots with a mid value
	;
	movlw	0x7F
	movwf	BUFFER
	movwf	BUFFER+1

	movlw   0x03        ; Send report to PC via USB
	btfsc   CHANGED     ; Skip sending data if gamepad
	call    SendData    ;  hasn't been touched.

	;
	; Get LED states from the host.
	;
	bankisel BUFFER		; point IRP+FSR to destination buffer
	movlw	BUFFER
	movwf	FSR
	pagesel	GetEP1		; ensure PCLATH bits are set
	call	GetEP1
	pagesel	Loop
	btfss	STATUS,C	; was any data copied?
	goto	Loop		; no: keep trying

	;
	; Assume one byte read.  Put LED states onto port A
	;
	banksel	BUFFER
	movfw	BUFFER
	banksel	PORTA
	movwf	PORTA

	goto    Loop        ; Start loop all over again

;*******************************************************************
;SendData
; Sends data to the PC via USB.  The number of bytes to be sent is 
; set in W.  Data to be sent is found in registers starting with
; BUFFER.
;
; Input: W, BUFFER to BUFFER + (W-1)
;
;*******************************************************************
SendData
	movwf   TEMP
	pagesel	PutEP1          
	banksel BUFFER
	bcf	STATUS,IRP  ; Set FSR to point to Bank 0
	movlw	BUFFER      
	movwf	FSR 
	movf    TEMP, w	    ; number of bytes to be sent to PC
	call	PutEP1      
	pagesel Loop
	btfss	STATUS,C    ; loop here until packet goes through
	goto	SendData
	bcf     CHANGED
	return

;*******************************************************************
;GetData
; Gets data from the PC via USB.  The number of bytes to acquire is 
; set in W.  Data is received into registers starting with
; BUFFER.
;
; Input: W, BUFFER to BUFFER + (W-1)
;
;*******************************************************************
;GetData
;	movwf   TEMP		; number of bytes
;	banksel BUFFER
;	pagesel	GetEP1
;	movlw	BUFFER
;	movwf	FSR
;	movf    TEMP,w		; number of bytes to be sent to PC
;	call	PutEP1      
;	pagesel Loop
;	btfss	STATUS,C	; loop here until packet goes through
;	goto	SendData
;	bcf     CHANGED
;	return

;*******************************************************************
;ReadButtons
; Reads the state of the buttons and places the result in W.
;   
; Output: W
;*******************************************************************
ReadButtons
	banksel BUTTONS		; Must be same bank as PORTB
	clrf    BUTTONS		; This is the buffer the button states
				;   are sent in. 
	btfss   PORTB,7     
	bsf     BUTTONS,0	; BUTTON 1
	btfss   PORTB,6
	bsf     BUTTONS,1	; BUTTON 2
	btfss   PORTB,5
	bsf     BUTTONS,2	; BUTTON 3
	btfss   PORTB,4
	bsf     BUTTONS,3	; BUTTON 4

	movf    BUTTONS,w
	return

;*******************************************************************
;InitGameport
; Initialize applicable registers of the PICmicro.    
;*******************************************************************
InitGameport
	;
	; Clear output port latches.
	;
	banksel	PORTA
	clrf	PORTA
	clrf    PORTB
	clrf	PORTC

	;
	; Set port directions.
	;
	banksel TRISB

	;
	; Port A:
	;
	; Set to all outputs
	;
	clrf	TRISA & 0x7F

	;
	; Port B:
	;
	; Pins 7..4 correspond to buttons 1..4, so need to be
	; inputs.  Pins 3..0 are outputs.
	;
	movlw	B'11110000'
	movwf	TRISB

	;
	; ADCON1
	;
	; Disable analogue inputs so that port A's outputs
	; can be read back correctly.  The device powers up in such
	; a way that analogue inputs read back as 0.
	;
	movlw	(1<<PCFG2) | (1<<PCFG1) | (1<<PCFG0)	; all digital
	movwf	ADCON1 & 0x7F

	;
	; OPTION_REG
	;
	movlw   B'00000000'
	movwf   OPTION_REG ^ 0x80 ; enable pullups on port B

	;
	; ADCON0
	;
	banksel ADCON0
	clrf    ADCON0

	return

	end

